////////////////////////////////////////////////////
//                                                //
// bz2tag -- a tool to convert blastz output      //
//           to tag format used by homologMiner   //
// ver 1.02                                       //  
//                                                //   
// Language: standard C                           // 
// Platform: Linux Debian                         //
// Contact:  Minmei Hou, PhD Candidate            //  
//           CSE Dept., Penn State University     //
//           mhou@cse.psu.edu                     //    
////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
//  Build Process                                            //
///////////////////////////////////////////////////////////////
//  Required files:                                          //
//    bz2tag.c                                               //
//                                                           //
//  compiler command:                                        //
//    gcc -g -Wall bz2tag.c -o bz2tag                        //
///////////////////////////////////////////////////////////////


/*
  Maintenance History:
  ====================
  ver 1.02: 02 Jan 2007
  - added error checking for chromosome number
  ver 1.01: 25 Dec 2006
  - removed usage of util.h & util.c
  ver 1.00: 17 Nov 2006
  - first release
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct bzTag{
  struct bzTag* next;
  int chr1, chr2;
  int b1, e1, b2, e2, pci;
  char strand2;
};

#define MAX_ALIGN_LEN 15000000

int bigger_int(int a, int b) {
  return ( a > b ? a : b);
}



int main(int argc, char **argv) {
  char buf[500], chr1[500], chr2[500],  name1[200], name2[200], strand1, strand2, cur_file1[500], cur_file2[500];
  int k, b1, b2, e1, e2, old_e1, old_e2, dir1,dir2, contig1, contig2;
  int srcSize1, srcSize2, tmp, start1, start2, end1, end2, pci;
  int seqstart1, seqstart2, seqend1, seqend2, totalpci, totalseg;
  FILE *fp;
  struct bzTag tag;
  
  if (argc < 2) {
    fprintf(stderr, " -- convert blastz output to tag file.\n args: blastz-output\n");
    exit(1);
  }
  
  fp = fopen(argv[1], "r");
  if ( (fgets(buf, 500, fp) == NULL) || (strcmp(buf, "#:lav\n")!=0) ) {
    fprintf(stderr, "%s is not a blastz output file", argv[1]);
    exit(1);
  }
  while (fgets(buf, 500, fp) && (strcmp(buf, "#:lav\n")!=0) )
    if ( strcmp(buf, "d {\n")==0 ) {
      printf("#\n");
      while (fgets(buf, 500, fp) && buf[0] != '}')
	printf("#%s", buf+1);
    }
  tag.next = NULL;
  while (fgets(buf, 500, fp)) {
    if ( strcmp(buf, "s {\n")==0 ) {
      fgets(buf, 500, fp);
      printf("# %s", buf);
      if (sscanf(buf, "  \"%s %d %d %d %d", cur_file1, &seqstart1, &seqend1, &dir1, &contig1) != 5) {	
	fprintf(stderr, "Wrong format, cannot find seq file or orient or contig in %s\nPlease see README file\n", buf);
	exit(1);
      }
      contig1--;
      tmp = strlen(cur_file1);
      if (dir1 == 0) 
	cur_file1[tmp-1] = '\0';
      else
	cur_file1[tmp-2]='\0';
      
      fgets(buf, 500, fp);
      printf("# %s", buf);
      if (sscanf(buf, "  \"%s %d %d %d %d", cur_file2, &seqstart2, &seqend2, &dir2, &contig2) != 5) {
	fprintf(stderr, "Wrong format, Cannot find seq file or orient or contig in %s\nPlease see README file\n", buf);
	exit(1);
      }
      contig2--;
      tmp = strlen(cur_file2);
      if ( dir2 == 0)
	cur_file2[tmp-1]='\0';
      else 
	cur_file2[tmp-2]='\0';

    }
    else if ( strcmp(buf, "h {\n")==0 ) {      
      fgets(buf, 500, fp);
      printf("# %s", buf);
      if ( sscanf(buf, "  \"%[^:]:%[^:]:%d-%d:%c:%d", name1, chr1, &start1, &end1, &strand1, &srcSize1) != 6 ) {
	fprintf(stderr, "Wrong format: %s, please see README file.\n", buf);
	exit(1);
      }
      fgets(buf, 500, fp);
      printf("# %s", buf);
      if ( sscanf(buf, "  \"%[^:]:%[^:]:%d-%d:%c:%d", name2, chr2, &start2, &end2, &strand2, &srcSize2) != 6 ) {
	fprintf(stderr, "Wrong format: %s, please see README file.\n", buf);
	exit(1);
      }
      
      //need to change strand according to dir1,2, since header didn't change
      if ( dir2 == 0 )
	tag.strand2 = '+';
      else
	tag.strand2 = '-';

/*
      if ( chr1[3] == 'X' || chr1[3] == 'x' )
	tag.chr1 = 23;
      else if ( chr1[3] == 'Y' || chr1[3] == 'y' )
	tag.chr1 = 24;
	*/
	if ( chr1[3] < '0' || chr1[3] > '9' )
		fprintf(stderr, "Wrong format: %s\nPlease see README file.\n", chr1);
  else
		tag.chr1 = atoi(chr1+3);
      
/*      
      if ( chr2[3] == 'X' || chr2[3] == 'x' )
	tag.chr2 = 23;
      else if ( chr2[3] == 'Y' || chr2[3] == 'y' )
	tag.chr2 = 24;
	*/
	if ( chr2[3] < '0' || chr2[3] > '9' )
		fprintf(stderr, "Wrong format: %s\nPlease see README file.\n", chr2);
  else
		tag.chr2 = atoi(chr2+3);
      
      //tag.strand2 = strand2; // assume the strand1 is +      
    }			  
    else if ( strcmp(buf, "a {\n")==0 ) {
      tag.b1 = tag.e1 = tag.b2 = tag.e2 = -1;
      tag.pci = -1;
      fgets(buf, 500, fp); // score
      fgets(buf, 500, fp); // can't trust this b1, b2;
      sscanf(buf, "  b %d %d", &b1, &b2);
      fgets(buf, 500, fp); // ditto for e1, e2
      sscanf(buf, "  e %d %d", &e1, &e2);
      /*
      if (sscanf(buf, "  e %d %d", &e1, &e2) != 2) {
	fprintf(stderr, "cannot parse: %s", buf);
	exit(1);
      }
      */
      old_e1 = old_e2 = -1;
      k = 0;
      //while (fgets(buf, 500, fp) && buf[0] != '}') {
      //if (sscanf(buf, "  l %d %d %d %d %d", &b1, &b2, &e1, &e2, &pci) != 5) {
      //  fprintf(stderr, "cannot read end-points: %s", buf);
      //  exit(1);
      //	}
      if (old_e1 == -1) {
	tag.b1 = start1 + b1 - 1;
	if ( tag.strand2 == '+')
	  tag.b2 = start2 + b2 - 1;
	else
	  tag.b2 = end2 - e2;
      } else {
	if (k + bigger_int(e1+start1-1-tag.b1,e2+start2-1-tag.b2) >= MAX_ALIGN_LEN) {
	  fprintf(stderr, "bz2tag: alignment too long.\n");
	  exit(1);
	}	  
      }
      old_e1 = e1;
      old_e2 = e2;
      //}
      tag.e1 = start1 + e1 - 1;
      if ( tag.strand2 == '+')
	tag.e2 = start2 + e2 - 1;
      else
	tag.e2 = end2 - b2;
      //tag.pci = pci;
      totalseg=totalpci=0;
      while (fgets(buf, 500, fp) && buf[0] != '}') {
	if (sscanf(buf, "  l %*d %*d %*d %*d %d", &pci) != 1) {
	  fprintf(stderr, "cannot read end-points: %s", buf);
	  exit(1);
	}     
	totalpci += pci;
	totalseg++;
      }

      tag.pci = totalpci/totalseg;
      
      if ( tag.b1 == tag.b2 && tag.e1 == tag.e2 && tag.strand2 == '+' && tag.chr1 == tag.chr2)
	continue;
      printf("t %d %d %d %d %c %d %d %d\n", tag.chr1, tag.b1, tag.e1, tag.chr2, tag.strand2, tag.b2, tag.e2, tag.pci);
      //if ( self == 0 )  // add the contrary
      //printf("t %d %d %d %d %c %d %d %d\n", tag.chr2, tag.b2, tag.e2, tag.chr1, tag.strand2, tag.b1, tag.e1, tag.pci);
    }
  }
  
  return 0;
}
