////////////////////////////////////////////////////////
//                                                    //
// bzTag.cpp -- this file contains implementation     //
//           of tag structure and operations          // 
// ver 1.01                                           //  
//                                                    //   
// Language: standard C++                             // 
// Platform: Linux Debian                             //
// Contact:  Minmei Hou, PhD Candidate                //  
//           CSE Dept., Penn State University         //
//           mhou@cse.psu.edu                         //    
////////////////////////////////////////////////////////


#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include "bzTag.hpp"

extern short MAX_NUM_CHRS;
extern int CHRS_SEPERATE;

long *fakestarts;
int* init_chrs;

bool compar_tag_b1(bzTag* a, bzTag* b) {
  return a->b1 < b->b1;
}

void initialize_chrs(const char* tagFileName) {
  char inputline[1000];
  ifstream tagFile;
  istringstream is;
  bzTag tag;
  int i, num_chrs, j, prevsize, currsize;
  char ch;

  init_chrs = (int*)malloc((MAX_NUM_CHRS+1)*sizeof(int));
  memset(init_chrs, 0, MAX_NUM_CHRS+1);

  tagFile.open(tagFileName);
  while ( tagFile ) {
    tagFile.getline(inputline, 1000);
    if ( inputline[0] != 't' )
      continue;
    is.clear();
    is.str(string(inputline));
    is >> ch >> tag.chr1 >> tag.b1 >> tag.e1 >> tag.chr2
       >> tag.strand2 >> tag.b2 >> tag.e2 >> tag.pci;
    if ( tag.chr1 > MAX_NUM_CHRS || tag.chr2 > MAX_NUM_CHRS ) {
      fprintf(stderr, "the chromosome index exceeds maximum: %s or %s\n", tag.chr1, tag.chr2);
      exit(1);
    }
    if ( tag.e1 > init_chrs[tag.chr1] )
      init_chrs[tag.chr1] = tag.e1;
    if ( tag.e2 > init_chrs[tag.chr2] )
      init_chrs[tag.chr2] = tag.e2;    
  }
  tagFile.close();

  num_chrs=0;
  for (i=1; i<=MAX_NUM_CHRS; i++)
    if ( init_chrs[i] > 0 )
      num_chrs++;
  
  fakestarts = (long*)malloc(num_chrs*sizeof(long));
  for (i=1, j=0; i<=MAX_NUM_CHRS; i++)
    if ( init_chrs[i] > 0 ) {
      fakestarts[j] = init_chrs[i];
      init_chrs[i] = j;
      j++;
    }
  
  currsize = fakestarts[0];
  fakestarts[0] = 0;
  for (j=1; j<num_chrs; j++) {
    prevsize = currsize;
    currsize = fakestarts[j];
    fakestarts[j] = fakestarts[j-1] + prevsize + CHRS_SEPERATE;
  }
}

void free_chrs() {
  free(fakestarts);
  free(init_chrs);
}

//-------< read from tag-file-name >---------
//-------- return sorted vector of tags -----
vector<bzTag*>* readTagFile(const char* tagFileName) {
  char inputline[1000];
  vector<bzTag*>* tags = new vector<bzTag*>;
  ifstream tagFile;
  istringstream is;
  bzTag *tag;
  char ch;

  tagFile.open(tagFileName);
  while ( tagFile ) {
    tagFile.getline(inputline, 1000);
    if ( inputline[0] != 't' )
      continue;
    is.clear();
    is.str(string(inputline));
    tag = new bzTag;
    is >> ch >> tag->chr1 >> tag->b1 >> tag->e1
       >> tag->chr2 >> tag->strand2 >> tag->b2
       >> tag->e2 >> tag->pci;
    tag->b1 += fakestarts[init_chrs[tag->chr1]];
    tag->e1 += fakestarts[init_chrs[tag->chr1]];
    tag->b2 += fakestarts[init_chrs[tag->chr2]];
    tag->e2 += fakestarts[init_chrs[tag->chr2]];
    tags->push_back(tag);
  }
  tagFile.close();
  sort(tags->begin(), tags->end(), compar_tag_b1);
  return tags;
}

void printTag(bzTag* tag, ostream& os) {
  os << "t " << tag->chr1 
     << ' ' << tag->b1 - fakestarts[init_chrs[tag->chr1]]
     << ' ' << tag->e1 - fakestarts[init_chrs[tag->chr1]]
     << ' ' << tag->chr2 << ' ' << tag->strand2
     << ' ' << tag->b2 - fakestarts[init_chrs[tag->chr2]]
     << ' ' << tag->e2 - fakestarts[init_chrs[tag->chr2]]
     << ' ' << tag->pci << std::endl;
}
