/////////////////////////////////////////////////////////////
//                                                         //
// graph_flow_min_cut_ll.cpp -- this file contains         //
//           implementation of Ford-Fulkerson algorithm    //
// ver 1.01                                                //  
//                                                         //   
// Language: standard C++                                  // 
// Platform: Linux Debian                                  //
// Contact:  Minmei Hou, PhD Candidate                     //  
//           CSE Dept., Penn State University              //
//           mhou@cse.psu.edu                              //    
/////////////////////////////////////////////////////////////


#include <deque>
#include <list>
#include <iostream>
#include <iterator>
#include "graph_node.hpp"
#include "graph_flow_min_cut_ll.hpp"

using namespace std;

vector< list<flow_edge*>* >* flow_graph;
char *markArr1, *markArr2, *visitArr, *markArr, *ccArr;
int *parentArr;

void initialize_flow_graph(vector<node*>& nodes) {
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  list<flow_edge*>* nlist, *nlist2;
  list<flow_edge*>::iterator listIter1, listIter2;
  flow_edge* nedge;
  int nodesize=nodes.size(), i, node_index2;

  flow_graph = new vector< list<flow_edge*>* >;

  for (i=0; i<nodesize; i++) {
    nlist = new list<flow_edge*>;
    flow_graph->push_back(nlist);
    edges = nodes[i]->edges;
    //edgesize = edges->size();
    for (edgeIt = edges->begin(); edgeIt != edges->end(); ++edgeIt ) {
      nedge = new flow_edge;
      nedge->target_index = (*edgeIt)->node_index;
      nedge->Gc = nedge->Rc = 1;
      nedge->Gf = 0;
      nedge->reverse = NULL;
      nlist->push_back(nedge);
    }
  }
  
  for (i=0; i<nodesize; i++) {
    nlist = (*flow_graph)[i];
    for (listIter1 = nlist->begin(); listIter1 != nlist->end(); ++listIter1) {
      if ( (*listIter1)->reverse != NULL )
	continue;
      node_index2 = (*listIter1)->target_index;
      nlist2 = (*flow_graph)[node_index2];
      for (listIter2 = nlist2->begin(); listIter2 != nlist2->end(); ++listIter2)
	if ( (*listIter2)->target_index == i )
	  break;
      if (listIter2 == nlist2->end()) {
	cerr << "end of list\n";
	continue;
      }
      else {
	(*listIter1)->reverse = *listIter2;
	(*listIter2)->reverse = *listIter1;
      }
    }
  }

  markArr1 = (char*)malloc(nodesize*sizeof(char));
  markArr2 = (char*)malloc(nodesize*sizeof(char));
  markArr  = (char*)malloc(nodesize*sizeof(char));
  visitArr = (char*)malloc(nodesize*sizeof(char));
  parentArr= (int*)malloc(nodesize*sizeof(int));
}

void delete_flow_graph() {
  list<flow_edge*>::iterator listIter;
  list<flow_edge*>* nlist;
  int nodesize = flow_graph->size(), i;
  
  for (i=0; i<nodesize; i++) {
    nlist = (*flow_graph)[i];
    for (listIter=nlist->begin(); listIter != nlist->end(); ++listIter)
      delete (*listIter);
    nlist->clear();
    delete nlist;
  }
  flow_graph->clear();
  delete flow_graph;
  free(markArr1);
  free(markArr2);
  free(markArr);
  free(visitArr);
  free(parentArr);
}

int consistent(vector<node*>& nodes, deque<int>& pathqueue, int);

// return the shortest path between i and j
// return 0: no path, 1: consistent, 2: non-consistent
int breadth_first_shortest_path(vector<node*>& nodes, int i, int j) {
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  int nodeSize=nodes.size(), k, curr;
  deque<int> walkqueue;
  
  for (k=0; k<nodeSize; k++) {
    nodes[k]->visited = 'n';
    nodes[k]->parent = -1;
  }
  
  walkqueue.push_back(i);
  nodes[i]->visited = 'y';
  while ( !walkqueue.empty() ) {
    curr = walkqueue.front();
    walkqueue.pop_front();
    edges = nodes[curr]->edges;
    //edgeSize = edges->size();
    //for (k=0; k<edgeSize; k++)
    for (edgeIt = edges->begin(); edgeIt != edges->end(); ++edgeIt)
      if ( nodes[ (*edgeIt)->node_index ]->visited == 'n' ) {
	walkqueue.push_back( (*edgeIt)->node_index );
	nodes[(*edgeIt)->node_index]->visited = 'y';
	nodes[(*edgeIt)->node_index]->parent = curr;
	if ( (*edgeIt)->node_index == j )
	  break;
      }
    if ( edgeIt != edges->end() )
      break;
  }
  if ( walkqueue.empty() ) {
    cerr << "No path exist. ";
    return 0;
  }
  
  deque<int> pathqueue;
  int queueSize;
  k=j;
  while ( k != -1 ) {
    pathqueue.push_front(k);
    k = nodes[k]->parent;
  }
  queueSize = pathqueue.size();
  //cerr << "(path: ";
  //for (k=0; k<queueSize; k++)
  //cerr << pathqueue[k] << ' ';
  int cons = consistent(nodes, pathqueue, 50);
  //cerr << cons << ") ";
  if ( cons == -1 )
    return 1;
  else
    return 2;
}

void allocate_ccArr(int nodeSize) {
  ccArr = (char*)malloc(nodeSize*sizeof(char));
}

void free_ccArr() {
  free(ccArr);
}

void initialize_ccArr(int nodeSize, vector<int>* cc) {
  int k, ccSize=cc->size();
  
  memset(ccArr, 'n', nodeSize);
  for (k=0; k<ccSize; k++)
    ccArr[(*cc)[k]] = 'y';
}

// return the number of pairs which are consistent
// breadth walk, only add consistent nodes
int breadth_first_walk_tree(vector<node*>& nodes, int i, vector<int>* cc) {
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  int nodeSize=nodes.size(), k, curr, consistent_count=0;
  deque<int> walkqueue;
  deque<int> pathqueue;
  
  for (k=0; k<nodeSize; k++) {
    nodes[k]->visited = 'n';
    nodes[k]->parent = -1;
  }

  walkqueue.push_back(i);
  nodes[i]->visited = 'y';
  while ( !walkqueue.empty() ) {
    curr = walkqueue.front();
    walkqueue.pop_front();
    edges = nodes[curr]->edges;
    for (edgeIt = edges->begin(); edgeIt != edges->end(); ++edgeIt) {
      if ( ccArr[ (*edgeIt)->node_index ] == 'n' )
	continue;
      if ( nodes[ (*edgeIt)->node_index ]->visited == 'n' ) {
	pathqueue.push_front( (*edgeIt)->node_index );
	k = curr;
	while ( k != -1 ) {
	  pathqueue.push_front( k );
	  k = nodes[k]->parent;
	}
	if ( consistent(nodes, pathqueue, 1) == -1 ) {
	  walkqueue.push_back( (*edgeIt)->node_index );
	  nodes[(*edgeIt)->node_index]->visited = 'y';
	  nodes[(*edgeIt)->node_index]->parent = curr;
	  consistent_count++;
	}
	pathqueue.clear();
      }
    }
  }
  return consistent_count;
}

long x2y_after(vector<bzTag*>* tags, long x) {
  vector<bzTag*>::iterator tagIt;
  
  tagIt=tags->begin();
  while ( tagIt != tags->end() && (*tagIt)->e1 < x )
    ++tagIt;
  
  if ( (*tagIt)->b1 <= x ) {
    if ( (*tagIt)->strand2 == '+' )
      return (*tagIt)->b2 + x - (*tagIt)->b1;
    else
      return (*tagIt)->e2 - (x - (*tagIt)->b1 );
  }
  else {
    if ( (*tagIt)->strand2 == '+' )
      return (*tagIt)->b2;
    else
      return (*tagIt)->e2;
  }
}

long x2y_before(vector<bzTag*>* tags, long x) {
  int tagSize=tags->size(), k;
  vector<bzTag*>::iterator tagIt;

  k=tagSize-1;
  //tagIt = tags->end() - 1;
  tagIt = tags->end();
  tagIt--;
  while ( k>=0 && (*tagIt)->b1 > x) {
    k--;
    --tagIt;
  }
  if ( (*tagIt)->e1 >= x ) {
    if ( (*tagIt)->strand2 == '+' ) 
      return (*tagIt)->b2 + x - (*tagIt)->b1;
    else
      return (*tagIt)->e2 - (x - (*tagIt)->b1 );
  }
  else {
    if ( (*tagIt)->strand2 == '+' )
      return (*tagIt)->e2;
    else
      return (*tagIt)->b2;
  }
}

// assume tags are properly sorted in edges
// are edges along this path consistent or not, if not, ret the one not consistent
int consistent(vector<node*>& nodes, deque<int>& pathqueue, int consistency_overlap) {
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  vector<bzTag*>* tags;
  vector<bzTag*>::iterator tagIt;
  int pathSize=pathqueue.size(), k, node1, node2, node3, edgeSize;
  long x1, x2, y1, y2;

  x1 = x2 = -1;
  for (k=0; k<pathSize-2; k++) {
    node1 = pathqueue[k];
    node2 = pathqueue[k+1];
    node3 = pathqueue[k+2];
    edges = nodes[node1]->edges;
    edgeSize = edges->size();
    for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt)
      if ( (*edgeIt)->node_index == node2 )
	break;
    tags = (*edgeIt)->tags;
    if ( (*(tags->begin()))->strand2 == '+' ) {
      if ( x1 == -1 ) {
	y1 = (*(tags->begin()))->b2;
	y2 =  (*(tags->end()-1))->e2;
      }
      else {
	y1 = x2y_after(tags, x1);
	y2 = x2y_before(tags, x2);
      }
    }
    else {
      if ( x1 == -1 ) {
	y1 = (*(tags->end()-1))->b2;
	y2 = (*(tags->begin()))->e2;
      }
      else {
	y1 = x2y_before(tags, x2);
	y2 = x2y_after(tags, x1);
      }
    }
    
    edges = nodes[node2]->edges;
    edgeSize = edges->size();
    for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt)
      if ( (*edgeIt)->node_index == node3 )
	break;
    tags = (*edgeIt)->tags;
    x1 = (*(tags->begin()))->b1;
    x2 = (*(tags->end()-1))->e1;
    if ( y1 >= x2-consistency_overlap || x1 >= y2-consistency_overlap ) { // non-overlap
      return node2;
    }
    else { // update x1, x2
      x1 = y1 > x1 ? y1 : x1;
      x2 = y2 < x2 ? y2 : x2;
    }
  }
  return -1;
}

void update_flow_graph(vector< list<flow_edge*>* >* flow_graph, deque<int>* path) {
  list<flow_edge*>* nlist;
  list<flow_edge*>::iterator listIter;
  vector<flow_edge*> pathEdges;
  int pathsize = path->size();
  int i;
  short minc=10000; 
  
  for (i=0; i<pathsize-1; i++) {  
    nlist = (*flow_graph)[(*path)[i]];
    for (listIter=nlist->begin(); listIter != nlist->end(); ++listIter)
      if ( (*listIter)->target_index == (*path)[i+1] )
	break;
    if ( listIter == nlist->end() ) {
      cerr << "edge does not exist\n";
      return;
    }
    pathEdges.push_back(*listIter);
    if ( (*listIter)->Rc < minc )
      minc = (*listIter)->Rc;
  }
  for (i=0; i<pathsize-1; i++) {
    if ( pathEdges[i]->reverse->Gf > 0 ) {
      if ( pathEdges[i]->reverse->Gf >= minc )
        pathEdges[i]->reverse->Gf -= minc;
      else {
        pathEdges[i]->Gf += (minc - pathEdges[i]->reverse->Gf);
	pathEdges[i]->reverse->Gf = 0;
      }
    }
    else {
      pathEdges[i]->Gf += minc;
    }
  }
  for (i=0; i<pathsize-1; i++) 
    if ( pathEdges[i]->Gf > 0 ) {
      pathEdges[i]->Rc = pathEdges[i]->Gc - pathEdges[i]->Gf;
      pathEdges[i]->reverse->Rc = pathEdges[i]->reverse->Gc + pathEdges[i]->Gf;
    }
  pathEdges.clear();
}

// return the shortest path between i and j
deque<int>* breadth_shortest_path(int source, int target, vector<int>* part) {
  int  k, curr, size=flow_graph->size(), nodeIndex2;
  deque<int> walkqueue;
  list<flow_edge*>* nlist;
  list<flow_edge*>::iterator listIter;

  for (k=0; k<size; k++) {
    visitArr[k] = 'n';
    parentArr[k] = -1;
  }
  part->clear();
  walkqueue.push_back(source);
  part->push_back(source);
  visitArr[source] = 'y';
  while ( !walkqueue.empty() ) {
    curr = walkqueue.front();
    walkqueue.pop_front();
    nlist = (*flow_graph)[curr];
    for (listIter = nlist->begin(); listIter != nlist->end(); listIter++) {
      nodeIndex2 = (*listIter)->target_index;
      if ((*listIter)->Rc > 0 && visitArr[nodeIndex2] == 'n') {
	walkqueue.push_back( nodeIndex2 );
	part->push_back(nodeIndex2);
	visitArr[nodeIndex2] = 'y';
	parentArr[nodeIndex2] = curr;
	if ( nodeIndex2 == target )
	  break;
      }
    }
    if ( listIter != nlist->end() )
      break;
  }
  if ( walkqueue.empty() ) {
    cerr << "No path exist\n";
    return NULL;
  }
  
  deque<int>* pathqueue = new deque<int>;
  k=target;
  while ( k != -1 ) {
    pathqueue->push_front(k);
    k = parentArr[k];
  }
  part->clear();
  return pathqueue;
}

vector<int>* Ford_Fulkerson(vector<int>* cc, int source, int target) {
  int i, ccSize=cc->size();
  deque<int>* pathqueue;
  vector<int>* part = new vector<int>;
  list<flow_edge*>* nlist;
  list<flow_edge*>::iterator listIter;

  for (i=0; i<ccSize; i++) {
    nlist = (*flow_graph)[(*cc)[i]];
    for (listIter=nlist->begin(); listIter != nlist->end(); ++listIter) {
      (*listIter)->Gc = (*listIter)->Rc = 1;
      (*listIter)->Gf = 0;
    }
  }
  
  pathqueue = breadth_shortest_path(source, target, part);
  while ( pathqueue != NULL) {
    update_flow_graph(flow_graph, pathqueue);
    pathqueue->clear();
    delete pathqueue;
    pathqueue = breadth_shortest_path(source, target, part);
  }
  delete pathqueue;
  return part;
}

vector<int>* component_complement(int nodesize, vector<int>* cc, vector<int>* component) { 
  int ccsize=cc->size(), i, componentsize=component->size(); 
  vector<int>* complement = new vector<int>; 
   
  memset(markArr, 'n', nodesize);
   
  for (i=0; i<componentsize; i++) // for all c1 elements
    markArr[ (*component)[i] ] = 'y'; 
 
  for (i=0; i<ccsize; i++) // the other part of cc
    if ( markArr[ (*cc)[i] ] == 'n' ) 
      complement->push_back( (*cc)[i] ); 
  return complement; 
} 
 
vector< cut_edge* >* min_cut(vector<int>* c1, vector<int>* cc) { 
  int nodesize=flow_graph->size(), ccsize1=c1->size(), ccsize2, i, nodeIndex2; 
  cut_edge* ce; 
  vector< cut_edge*>* edge_pairs = new vector< cut_edge* >; 
  vector<int>* c2; 
  list<flow_edge*>* nlist;
  list<flow_edge*>::iterator listIter;

  memset(markArr1, 'n', nodesize);
  memset(markArr2, 'n', nodesize);
  for (i=0; i<ccsize1; i++)
    markArr1[ (*c1)[i] ] = 'y';
 
  c2 = component_complement(nodesize, cc, c1); 
  ccsize2 = c2->size(); 
  for (i=0; i<ccsize2; i++)
    markArr2[ (*c2)[i] ] = 'y';
 
  for (i=0; i<ccsize1; i++) {
    nlist = (*flow_graph)[(*c1)[i]];
    for (listIter=nlist->begin(); listIter != nlist->end(); ++listIter) {
      nodeIndex2 = (*listIter)->target_index;
      if ( markArr2[nodeIndex2] == 'y' ) {
	ce = new cut_edge;
	ce->node1 = (*c1)[i];
	ce->node2 = nodeIndex2;
	edge_pairs->push_back(ce);
	ce = new cut_edge;
	ce->node1 = nodeIndex2;
	ce->node2 = (*c1)[i];
	edge_pairs->push_back(ce);
      }
    }
  }
  c2->clear();
  delete c2;
  return edge_pairs; 
} 

void cut_flow_graph(vector<cut_edge*>* cut_edges) {
  int cutsize=cut_edges->size(), edgesize, i, node1, node2;
  list<flow_edge*>* edges;
  list<flow_edge*>::iterator edgeIt;

  for (i=0; i<cutsize; i++){ 
    node1 = (*cut_edges)[i]->node1;
    node2 = (*cut_edges)[i]->node2;
    edges = (*flow_graph)[node1];
    edgesize = edges->size();
    for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt)
      if ( (*edgeIt)->target_index == node2 )
	break;
    if ( edgeIt != edges->end() ) {
      delete *edgeIt;
      edges->erase( edgeIt );
    }
    else
      cerr << "flow_graph Error!\n";
  }
}


vector<cut_edge*>* max_flow_min_cut(vector<node*>& nodes, vector<int>* cc, int i, int j) {
  vector<int>* part;
  int k, partsize;
  vector<cut_edge*>* cut_edges;

  if ( nodes[i]->edges->size() < nodes[j]->edges->size() )
    part = Ford_Fulkerson(cc, i, j);
  else
    part = Ford_Fulkerson(cc, j, i);

  // no path between source adn target on Rc, two parts:
  partsize = part->size();
  cerr << "(part: ";
  for(k=0; k<partsize; k++) {
    cerr << (*part)[k]<< ' ';
  }
  cerr << ") ";

  cut_edges = min_cut(part, cc);
  part->clear();
  delete part;
  cut_flow_graph(cut_edges);
  return cut_edges;
}
