/////////////////////////////////////////////////////////////
//                                                         //
// graph_flow_min_cut_ll.hpp -- this file contains         //
//           structure definition, public interface of     //
//           Ford-Fulkerson algorithm                      //
// ver 1.01                                                //  
//                                                         //   
// Language: standard C++                                  // 
// Platform: Linux Debian                                  //
// Contact:  Minmei Hou, PhD Candidate                     //  
//           CSE Dept., Penn State University              //
//           mhou@cse.psu.edu                              //    
/////////////////////////////////////////////////////////////

#ifndef GRAPH_FLOW_MIN_CUT_LL_HPP
#define GRAPH_FLOW_MIN_CUT_LL_HPP

#include <vector>
#include "graph_node.hpp"

struct flow_edge {
  int target_index;
  short Gc, Rc, Gf;
  flow_edge* reverse;
};

void initialize_flow_graph(vector<node*>& nodes);
void delete_flow_graph();
vector<cut_edge*>* max_flow_min_cut(vector<node*>& nodes, vector<int>* cc, int i, int j);
int breadth_first_shortest_path(vector<node*>& nodes, int i, int j);
int breadth_first_walk_tree(vector<node*>& nodes, int i, vector<int>* cc);
int consistent(vector<node*>& nodes, deque<int>& pathqueue, int consistency_overlap);

void allocate_ccArr(int nodeSize);
void free_ccArr();
void initialize_ccArr(int nodeSize, vector<int>* cc);

#endif
