//input: edge-file, ccs-file, the index of cc needs to be processed
// read in edge-file and ccs-file, transform the cc specified by argument into a graph with index 0, 1, 2..., and call minCut, and transform back

#include <vector>
#include <iostream>
#include <fstream>
#include "nodeedge.hpp"
#include "minCut.hpp"

using namespace std;

int *origGraph2cutGraph;
int *cutGraph2origGraph;


// transform the cc to the graph required by minCut
void produceMapping(vector<int>* cc, vector< vector<edge*>* >* orignodes){
  int i, origGraphSize, cutGraphSize;
  origGraphSize = orignodes->size();
  cutGraphSize = cc->size();
  origGraph2cutGraph = (int*)malloc(origGraphSize*sizeof(int));
  cutGraph2origGraph = (int*)malloc(cutGraphSize*sizeof(int));
  for (i=0; i<origGraphSize; i++)
    origGraph2cutGraph[i] = -1;
  for (i=0; i<cutGraphSize; i++)
    cutGraph2origGraph[i] = -1;
  for (i=0; i<cutGraphSize; i++) {
    cutGraph2origGraph[i] = (*cc)[i];
    origGraph2cutGraph[ (*cc)[i] ] = i;
  }
}

vector<node*>* transformGraph(vector<int>* cc, vector< vector<edge*>* >* orignodes) {
  vector<node*>* newgraph = new vector<node* >;
  vector<edge*> *origedges;
  node* newnode;
  edge *e, *newedge;
  int ccSize=cc->size(), i, edgeSize, j;

  for (i=0; i<ccSize; i++) {
    origedges = (*orignodes)[ (*cc)[i] ];
    edgeSize = origedges->size();
    newnode = new node;
    newnode->origNodes = new vector<int>;
    newnode->origNodes->push_back(i);
    newnode->edges = new vector<edge*>;
    for (j=0; j<edgeSize; j++) {
      e = (*origedges)[j];
      newedge = new edge(*e);
      newedge->index = origGraph2cutGraph[ e->index ];
      if ( newedge->index == -1 ) 
	delete newedge;
      else
	newnode->edges->push_back(newedge);
    }
    newgraph->push_back(newnode);
  }
  return newgraph;
}

void transformPartition(graphPartition* parts) {
  vector<int>* part;

  int size, i;

  part = parts->part1;
  size = part->size();
  for (i=0; i<size; i++)
    (*part)[i] = cutGraph2origGraph[ (*part)[i] ];
  part = parts->part2;
  size = part->size();
  for (i=0; i<size; i++)
    (*part)[i] = cutGraph2origGraph[ (*part)[i] ];
}

vector<int>* complement_partition(vector<int>* part1, int total_size) {
  vector<int>* part2 = new vector<int>;
  char *array = (char*)malloc(total_size*sizeof(char));
  int i, part1size = part1->size();
  for (i=0; i<total_size; i++)
    array[i] = 'u';
  for (i=0; i<part1size; i++)
    array[ (*part1)[i] ] = 'y';
  for (i=0; i<total_size; i++)
    if ( array[i] == 'u' )
      part2->push_back(i);
  return part2;
}
int main(int argc, char** argv) {
  vector< vector<edge*>* >* nodes;
  vector< vector<int>* >* ccs;
  vector<node*>* newgraph;
  graphPartition* parts;
  int ccs_index;

  if ( argc < 4 ) {
    cerr << "args: edge-file ccs-file ccs-index\n";
    return 1;
  }

  nodes = readedges(argv[1]);
  ccs   = readccs(argv[2]);
  ccs_index = atoi(argv[3]);

  (*ccs)[0];

  produceMapping((*ccs)[ccs_index], nodes);
  newgraph = transformGraph((*ccs)[ccs_index], nodes);
  parts = minCut(*newgraph);
  parts->part2 = complement_partition(parts->part1, (*ccs)[ccs_index]->size() );
  transformPartition(parts);
  cout << "0 : ";
  int i, partsize;
  partsize = parts->part1->size();
  cout << partsize << " : ";
  for (i=0; i<partsize; i++)
    cout << (*(parts->part1))[i] << ' ';
  cout << "\n1: ";
  partsize = parts->part2->size();
  cout << partsize << " : ";
  for (i=0; i<partsize; i++)
    cout << (*(parts->part2))[i] << ' ';
  cout << endl;

  return 0;
}
