/////////////////////////////////////////////////////////////
//                                                         //
// splitDupRepeat.cpp -- this file contains implementation //
//           of ANTI-EXCESS                                //
// ver 1.01                                                //  
//                                                         //   
// Language: standard C++                                  // 
// Platform: Linux Debian                                  //
// Contact:  Minmei Hou, PhD Candidate                     //  
//           CSE Dept., Penn State University              //
//           mhou@cse.psu.edu                              //    
/////////////////////////////////////////////////////////////


#include <iostream>
#include <fstream>
#include <stdlib.h>
#include "bzTag.hpp"
#include "splitDupRepeat.hpp"

extern int REPEAT_COPY_THRESHOLD;
extern int DUP_GAP;

bool compar_eventDot(eventDot *a, eventDot *b) {
  return a->pos < b->pos;
}

int determRepeat(vector<eventDot*>& dots, int count) {
  int max, copy, i;

  sort(dots.begin(), dots.end(), compar_eventDot);
  for (max=copy=i=0; i<count; i++) {
    if ( dots[i]->event == 'b' )
      copy++;
    else
      copy--;
    if ( copy > max )
      max = copy;
  }
  if ( max > REPEAT_COPY_THRESHOLD )
    return 1;
  else
    return 0;
}

// binary search
int searchRepeat(int* repeatFronts, int* repeatEnds, int count, int queryFront, int queryEnd) {
  int start, end, middle;

  start = 0;
  end = count-1;
  middle = end/2;
  while ( start <= end ) {
    middle = start + (end-start)/2;
    if ( queryFront > repeatEnds[middle] ) 
      start = middle+1;
    else if ( queryEnd < repeatFronts[middle] )
      end = middle-1;
    else
      return middle;
  }
  return -1;
}

vector<bzTag*>* splitDupRepeat(vector<bzTag*>* alltags, ofstream& ofdup, ofstream& ofrep) {
  vector<eventDot*> dots;
  eventDot* newdot;
  int count = 0, tagsize, i, j, n, interval, repeat, front_flag, end_flag, dup_index, rep_index;
  long *currEnds, min_pos, max_pos;
  vector<bzTag*>& tags = *alltags;
  vector<bzTag*>* dupTags = new vector<bzTag*>;

  min_pos = MAX_INT;
  max_pos = 0;

  tagsize = tags.size();
  for (i=0; i<tagsize; i++) {
    if ( tags[i]->b1 < min_pos )
      min_pos = tags[i]->b1;
    if ( tags[i]->b2 < min_pos )
      min_pos = tags[i]->b2;
    if ( tags[i]->e1 > max_pos )
      max_pos = tags[i]->e1;
    if ( tags[i]->e2 > max_pos )
      max_pos = tags[i]->e2;
  }
 
  currEnds = (long*)malloc(tagsize*sizeof(long));
  currEnds[0] = tags[0]->e1;
  for (i=1; i<tagsize; i++) {
    if ( tags[i]->e1 > currEnds[i-1] )
      currEnds[i] = tags[i]->e1;
    else
      currEnds[i] = currEnds[i-1];
  }

  dup_index = rep_index = 0;
  front_flag = 0;
  n=0;
  for (i=1; i<tagsize; i++) {
    if ( tags[i]->b1 - currEnds[i-1] > 500 || i==count-1) {
      if ( i == count - 1 )
				end_flag = i;
      else
				end_flag = i-1;
      interval = end_flag - front_flag + 1;
      for (j=0; j<interval; j++) {
				newdot = new eventDot;
				newdot->event = 'b';
				newdot->pos = tags[front_flag+j]->b2;
				dots.push_back(newdot);
				newdot = new eventDot;
				newdot->event = 'e';
        newdot->pos = tags[front_flag+j]->e2;
				dots.push_back(newdot);
      }
      
      repeat = determRepeat(dots, 2*interval);
      if ( repeat == 0 ) {
				ofdup << "c " << dup_index++ << endl;
				for (j=front_flag; j<=end_flag; j++) {
	  			printTag(tags[j], ofdup);	
	  			dupTags->push_back(tags[j]);
	  		}
      }
      else {
				ofrep << "c " << rep_index++ << endl;
				for (j=front_flag; j<=end_flag; j++)
	  			printTag(tags[j], ofrep);
      }
      
     	front_flag = i;
      for (j=0; j<2*interval; j++)
				delete dots[j];
      dots.clear();
    }
  }

	free(currEnds);
  return dupTags;
}

//----------< test entry point >----------------
//#define SPLIT_TEST
#ifdef SPLIT_TEST
int main (int argc, char** argv) {
  char filename[500];
  vector<bzTag*>* pTags, *dupTags;
  ofstream ofdup, ofrep, of;

  if ( argc < 2 ) {
    fprintf(stderr, "split a tag file into two: one has duplications, one has tandem repeat\nargs: tag-file\n");
    return 1;
  }

  pTags = readTagFile(argv[1]);
  
  sprintf(filename, "%s.dup", argv[1]);
  ofdup.open(filename);
  sprintf(filename, "%s.rep", argv[1]);
  ofrep.open(filename);
	
	dupTags = splitDupRepeat(pTags, ofdup, ofrep);
	ofdup.close();
	ofrep.close();
	
	int i, tagsize=dupTags->size();
	for (i=0; i<tagsize; i++) {
		printTag((*dupTags)[i], cout);
		delete (*dupTags)[i];
	}
	delete dupTags;
	
  /*
  vector<bzTag*>& tags = *ptags;

  min_pos = MAX_INT;
  max_pos = 0;

  tagsize = tags.size();
  for (i=0; i<tagsize; i++) {
    if ( tags[i]->b1 < min_pos )
      min_pos = tags[i]->b1;
    if ( tags[i]->b2 < min_pos )
      min_pos = tags[i]->b2;
    if ( tags[i]->e1 > max_pos )
      max_pos = tags[i]->e1;
    if ( tags[i]->e2 > max_pos )
      max_pos = tags[i]->e2;
  }

  repeatMarkArr = (char*)malloc( (max_pos-min_pos+1)*sizeof(char));
  for (k=0; k< (max_pos-min_pos+1); k++)
    repeatMarkArr[k] = 'n';
  
  currEnds = (long*)malloc(tagsize*sizeof(long));
  currEnds[0] = tags[0]->e1;
  for (i=1; i<tagsize; i++) {
    if ( tags[i]->e1 > currEnds[i-1] )
      currEnds[i] = tags[i]->e1;
    else
      currEnds[i] = currEnds[i-1];
  }
  
  dup_index = rep_index = 0;
  front_flag = 0;
  n=0;
  for (i=1; i<tagsize; i++) {
    if ( tags[i]->b1 - currEnds[i-1] > DUP_GAP || i==count-1) {
      if ( i == count - 1 )
	end_flag = i;
      else
	end_flag = i-1;
      interval = end_flag - front_flag + 1;
      for (j=0; j<interval; j++) {
	newdot = new eventDot;
	newdot->event = 'b';
	newdot->pos = tags[front_flag+j]->b2;
	dots.push_back(newdot);
	newdot = new eventDot;
	newdot->event = 'e';
        newdot->pos = tags[front_flag+j]->e2;
	dots.push_back(newdot);
      }
      
      repeat = determRepeat(dots, 2*interval);
      if ( repeat == 0 ) {
	ofdup << "c " << dup_index++ << endl;
	for (j=front_flag; j<=end_flag; j++)
	  printTag(tags[j], ofdup);	
      }
      else {
	ofrep << "c " << rep_index++ << endl;
	
	for (j=front_flag; j<=end_flag; j++) {
	  for (k=tags[j]->b1; k<=tags[j]->e1; k++)
	    repeatMarkArr[k - min_pos] = 'r';
	  for (k=tags[j]->b2; k<=tags[j]->e2; k++)
	    repeatMarkArr[k - min_pos] = 'r';
	}

	for (j=front_flag; j<=end_flag; j++)
	  printTag(tags[j], ofrep);
      }
      
      front_flag = i;
      for (j=0; j<2*interval; j++)
	delete dots[j];
      dots.clear();
    }
  }
  ofdup.close();
  ofrep.close();

  //----- below are commented --------------
  for (i=0, regionCount=0, inside=0; i<max_pos-min_pos+1; i++) {
    if ( inside==0 && repeatMarkArr[i] == 'r' ) {
      inside = 1;
      regionCount++;
    }
    if ( inside==1 && repeatMarkArr[i] == 'n' )
      inside = 0;
  }
  repeatFronts = (int*)malloc(regionCount*sizeof(int));
  repeatEnds   = (int*)malloc(regionCount*sizeof(int));
  for (i=j=inside=0; i<max_pos-min_pos+1; i++) {
    if ( inside==0 && repeatMarkArr[i] == 'r' ) {
      inside = 1;
      repeatFronts[j] = i+min_pos;
    }
    if ( inside == 1 && (repeatMarkArr[i] == 'n' || i==max_pos-min_pos) ) {
      inside = 0;
      repeatEnds[j] = i+min_pos;
      j++;
    }
  }
    
  sprintf(filename, "%s.dup2", argv[1]);
  fpwD = fopen(filename, "w");
  for (i=0; i<n; i++)
    if ( searchRepeat(repeatFronts, repeatEnds, regionCount, dupArr[i]->b1, dupArr[i]->e1)== -1 
	 && searchRepeat(repeatFronts, repeatEnds, regionCount, dupArr[i]->b2, dupArr[i]->e2)== -1 )
      fprintf(fpwD, "t2 %d %d %d %d %c %d %d %d\n", dupArr[i]->chr1, dupArr[i]->b1, dupArr[i]->e1, dupArr[i]->chr2, dupArr[i]->strand2, dupArr[i]->b2, dupArr[i]->e2, dupArr[i]->pci);
  fclose(fpwD);
  */
  
  return 0;
}
#endif
